// 				alignroot.C
// 	is_align_root_template()
//	is_align_root_distribute()
//
//  written by Zheng Gengbin

/*
This file contains 2 pass :
1. check in every scope if the align root is template .
   if not so , creat default template the same size as the array ,
   and align the array to this new template .
2. check in every scope if the align root(now be template) have been distributed .
   if not so , create a default processor and default distribute model .
   if the template is not used at all , delete the template from the scope .
these 2 passes are dependent on the align root information collected in info.C

		is_align_root_template()
		is_align_root_distribute()
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

#include "compilerApi.h"
#include "strategy.h"

#include "pass.h"
#include "ext_func.h"

extern int num_dims;

extern int silentFlag;

//inline int min(int x, int y) {return x < y ? x : y ; }
inline int max(int x, int y) {return x > y ? x : y ; }

/***************************************************************************
		is_align_root_template()
****************************************************************************/

// dimList is the size expr_list of array.
SgSymbol  *get_dim_template(SgStatement *hedr_stmt,int temp_dim,SgSymbol *sym,SgExpression *dimList)
{
	int variant = hedr_stmt->variant();
	ASSERT(variant == PROG_HEDR || variant == PROC_HEDR || variant == FUNC_HEDR  || variant == MODULE_STMT);

	  // expr_list will be the template dim created.
	SgExpression *expr_list=NULL;
	SgType *temptype;

	// if dimList is NULL , the size is the size of sym
	if (dimList==NULL)
	{
		SgArrayType *arraytype = isSgArrayType(sym->type());
		if (arraytype) {
			ASSERT(arraytype->getDimList());
			expr_list = arraytype->getDimList()->copyPtr();
		}
	}
	else {
		expr_list=dimList->copyPtr();
	}
          // the dimension of array
	int array_dim = expr_list?((SgExprListExp*)expr_list)->length():0;

	ASSERT(temp_dim >= array_dim);

	  // search if we already have such template defined before
	  // must also compare the distribute code
	SgStatement *dist_stmt1 = NULL;
	SgExpression *dist_op = NULL;
	SgExpression *procExp1 = NULL;
	if (sym->numberOfAttributes(Dist_Stmt) > 0)
        {	 // get distCode tree of symbol
		dist_stmt1=(SgStatement *)sym->attributeValue(0,Dist_Stmt);
		dist_op = dist_stmt1->expr(2)->lhs();
          	SgExpression *dist_target1 = dist_op->rhs();
          	procExp1 = dist_target1->rhs();
	}  // else the sym must be default distributed.

	  // check all the exist templates
	for (int i=0 ; i< hedr_stmt->numberOfAttributes(Temp_Def) ; i++)
	{
            SgExpression *dist_format1 = dist_op?dist_op->lhs()->rhs():(SgExpression *)NULL;
	    SgSymbol *tmp=(SgSymbol*)hedr_stmt->attributeValue(i,Temp_Def);
	      // ignore template with dynamic attribute 
	    if (tmp->numberOfAttributes(Dynamic_Stmt) > 0) continue;
	    SgType *tmptype = tmp->type();
	    int dim=0;
	    if (tmptype->variant() == T_ARRAY )
		dim = isSgArrayType(tmptype)->dimension();
	    if (dim == 0 && expr_list == NULL) return tmp;
//	    if (dim != ((SgExprListExp*)expr_list)->length() ) continue;
	    if (dim != temp_dim) continue;
	    int equal = 1;
	    // get distCode tree of template
	    if (tmp->numberOfAttributes(Dist_Stmt) == 0) 
	    {
		  // tmp must be default distributed, but also need to compare size.
		if (dist_stmt1 == NULL) 
		{
		    // sym must also be default distributed, but also need to compare size.
		    SgExpression *list = expr_list;		// distCode
		    SgExpression *dimlist = ((SgArrayType *)tmptype)->getDimList();  // type
		    for ( int j=0; j<array_dim && equal ; j++,list=list->rhs(),dimlist=dimlist->rhs() )
		    {
			// type compare
			if (*list->lhs() != *dimlist->lhs() ) 
				equal = 0;
		    }
		    // compare sub group processors
		    if (procExp1 && isSgArrayRefExp(procExp1)) equal = 0;
		}
		else
			equal = 0;
	    }
	    else
	    {
	   	SgStatement *dist_stmt2 = (SgStatement *)tmp->attributeValue(0,Dist_Stmt);
	   	SgExpression *dist_op = dist_stmt2->expr(2)->lhs();
          	SgExpression *dist_format2 = dist_op->lhs()->rhs();
          	SgExpression *dist_target2 = dist_op->rhs();
          	SgExpression *procExp2 = dist_target2->rhs();
		ASSERT(procExp2);
		// compare the array type and distribute pattern dimension by dimension
		SgExpression *list = expr_list;		// distCode
		SgExpression *dimlist = ((SgArrayType *)tmptype)->getDimList();  // type
		for (int j=0; j<array_dim && equal  ; j++,list=list->rhs(),dimlist=dimlist->rhs() )
		{
			  // type compare
			if (*list->lhs() != *dimlist->lhs() ) 
				equal = 0;
			  // distribute pattern
			if (dist_format1) 
			{
			    if ( dist_format1->lhs()->variant() != dist_format2->lhs()->variant() )
				equal = 0;
			    dist_format1=dist_format1->rhs();
			}
			else 	// default mapping
			{
#ifdef DEFAULT_COLLPASE
			    if (dist_format2->lhs()->variant() != STAR_OP)
				equal = 0;
#else
			    equal = 0;    		// not completed.
#endif
			}
			  // sub processor group
			if (procExp1 == NULL) {
			    if (isSgArrayRefExp(procExp2)) equal = 0;
			}
			else {
			    if (isSgArrayRefExp(procExp1) && isSgArrayRefExp(procExp2))
			    {
				if (*procExp1->lhs() != *procExp2->lhs()) equal = 0;
			    }
			}
			dist_format2=dist_format2->rhs();
		}	// end of for
	    }	// end of Dist_Stmt
//printf("Array:%s -- temp:%s\n",sym->identifier(),tmp->identifier());
	    if (equal) { return tmp; }
	} // end of for

	  // create new template
	SgSymbol *temp = get_temp_variable(hedr_stmt,"TMP");
	if (array_dim == temp_dim)
	{
		if (expr_list)
		   temptype = new SgArrayType(*SgTypeTemplates(),((SgExprListExp*)expr_list)->length(),*expr_list);
		else
		   temptype = SgTypeTemplates();
		temp->setType(temptype);
	}
	else if (array_dim < temp_dim) {
		ASSERT(array_dim>0);
		SgExprListExp *expr = isSgExprListExp(expr_list);
		ASSERT(expr);
		SgExpression *ele = expr->elem(array_dim-1);
		for (int i=array_dim+1; i<=temp_dim ; i++)
		{
			SgExpression *ele_copy = ele->copyPtr();
			expr->append(*ele_copy);
		}
		temptype = new SgArrayType(*SgTypeTemplates(),expr->length(),*expr_list);
	}

	temp->setType(temptype);

	SgExpression *type_op = new SgExpression(TYPE_OP,NULL,NULL,NULL,SgTypeTemplates());

	SgExpression *arrayref=new SgArrayRefExp(*temp,*expr_list);
	SgExpression *ref = new SgExpression(EXPR_LIST,arrayref,NULL,NULL); 

	SgStatement *decl=new SgStatement(TEMPLATE_STMT,NULL,NULL, ref,type_op,NULL);
        hedr_stmt->firstStmtForDeclaration()->insertStmtAfter(*decl, *hedr_stmt);
	
	hedr_stmt->addAttribute(Temp_Def , temp , sizeof(temp) );
	temp->addAttribute(Temp_Stmt , decl , sizeof(decl));
	return temp;
}

SgSymbol *add_align_to_template(SgStatement *hedr_stmt, SgSymbol *symbol)
{
	int variant = hedr_stmt->variant();
	ASSERT(variant == PROG_HEDR || variant == PROC_HEDR || variant == FUNC_HEDR || variant == MODULE_STMT);

	SgType *type=symbol->type();
	SgSymbol *temp;

	int dim=0;
	if (type->variant() == T_ARRAY)
		dim=((SgArrayType *)type)->dimension();

/*
	int nop = 0;
	if (symbol->numberOfAttributes(Dist_Stmt) > 0)
	{
printf("SYMBOL:%s\n",symbol->identifier());
	   // caculate * which will count nothing
	   SgStatement *dist_stmt=(SgStatement *)symbol->attributeValue(0,Dist_Stmt);
	   SgExpression *dist_op = dist_stmt->expr(2)->lhs();
           SgExpression *dist_format = dist_op->lhs()->rhs();
	   for (; dist_format ; dist_format=dist_format->rhs() )
	     if (dist_format->lhs()->variant() == STAR_OP) nop++;
	}
	printf("NOP:%d\n",nop);
*/

	int  temp_dim ;
	  // for example:   a(10); distribute a(block); P(2,2), 
	  // cannot write: distribute t(block) onto P
	  // default template's dim must large than default processor's
	if (dim < num_dims)
		temp_dim = num_dims;
	else 
		temp_dim = dim;

	if ( (symbol->attributes() &  ALLOCATABLE_BIT) == 0 )
		temp = get_dim_template(hedr_stmt,temp_dim,symbol,NULL);
	else {
		  // set size according to the alloc size
		SgExpression *list=NULL;
		for (int i=0 ; i< symbol->numberOfAttributes(Alloc_Size) ; i++)
		{
			list=(SgExpression *)symbol->attributeValue(i,Alloc_Size);
			if (list->symbol() == symbol ) break;
		}
		if (list == NULL) return NULL;
		 // lhs must be constant ??
		temp = get_dim_template(hedr_stmt,temp_dim,symbol,list->lhs());
	}

	  // now construct the new align statement
	SgExpression *array_expr_list=NULL;
	SgExpression *temp_expr_list=NULL;
	for (int i=max(dim,temp_dim)-1 ; i>=0; i--)
	{
		SgSymbol *dummy = get_temp_variable(hedr_stmt,"DM");
		SgVarRefExp *ref = new SgVarRefExp(dummy);
		if (i<dim)
		array_expr_list = new SgExpression(EXPR_LIST,ref,array_expr_list,NULL);
		if (i<temp_dim)
		temp_expr_list = new SgExpression(EXPR_LIST,ref,temp_expr_list,NULL);
	}
	// scalar not considered
	SgExpression *tempref = new SgArrayRefExp(*temp,*temp_expr_list);
	SgExpression *align_with = new SgExpression(ALIGN_WITH_OP,NULL,tempref,NULL);
	SgExpression *align_op = new SgExpression(ALIGN_OP,array_expr_list,align_with,NULL);
	SgExpression *expr_list = new SgExpression(EXPR_LIST , align_op ,NULL,NULL);

	SgExpression *varref = new SgVarRefExp(symbol);
	SgExpression *var_list = new SgExpression(EXPR_LIST, varref,NULL, NULL);

	SgStatement *align_decl;

	// insert statement
	if ( (symbol->attributes() &  ALLOCATABLE_BIT) == 0 )
	{
		align_decl=new SgStatement(ALIGN_DECL,NULL,NULL,var_list,NULL,expr_list);
		SgStatement *decl = symbol->declareInStmt(hedr_stmt);
		if (decl==NULL) decl=hedr_stmt->firstStmtForDeclaration();
        	decl->insertStmtAfter(*align_decl,*hedr_stmt);
	}
	else {
		// for allocatable symbol , add realign stmt to the
		for (int i=0 ; i< symbol->numberOfAttributes(Alloc_Stmt) ; i++)
		{
			align_decl=new SgStatement(REALIGN_STMT,NULL,NULL,var_list,NULL,expr_list);
			SgStatement *stmt=(SgStatement*)symbol->attributeValue(i,Alloc_Stmt);
        		stmt->insertStmtAfter(*align_decl,*stmt->controlParent());
		}
	}

	// add attributes
	temp->addAttribute(Align_Son , symbol , sizeof(symbol));
        symbol->addAttribute(Align_Parent , temp , sizeof(temp));
        symbol->addAttribute(Align_Stmt , align_decl , sizeof(align_decl));
	
	return temp;
}

// make sure if all the symbol align to a Template   ( if INHERIT , then no )
// dependence the  Align_Root collect !!
// must be   done before ALIGN_COLLAPSE_PASS
void is_align_root_template(SgStatement *hedr_stmt)
{
	for (int i=0 ; i< hedr_stmt->numberOfAttributes(Align_Root) ; i++)
        {
                SgSymbol *sym=(SgSymbol*)hedr_stmt->attributeValue(i,Align_Root);
		if (sym->type()->variant() == T_TEMPLATE) continue;
		if (((SgArrayType *)sym->type())->baseType()->variant() == T_TEMPLATE) continue;

		  // if have INHERIT attribute
		if (sym->attributes() & INHERIT_BIT) continue;

		  // if the symbol is a dummy , ignore it
		if ( sym->numberOfAttributes(DUMP_Def) != 0 )
			continue;

		//if (sym->numberOfAttributes(Dynamic_Stmt) != 0 ) continue;
		  // Allocatable
	        if (sym->attributes() &  ALLOCATABLE_BIT) continue;
		  // Pointer
	        if (sym->attributes() &  POINTER_BIT) continue;

if (!silentFlag)
printf("Align Root: %s is not a template \n",sym->identifier());

		SgSymbol *temp=add_align_to_template(hedr_stmt,sym);
		if (temp == NULL ) {
			printf("Error:failed align %s to default template .!\n",sym->identifier());
			return;
		}

		  // modified the distribute statement concerned about the sym
		if (sym->numberOfAttributes(Dist_Stmt) > 0)
		{
		   SgStatement *dist_stmt = (SgStatement *)sym->attributeValue(0,Dist_Stmt);
		      // if template already have been distributed
		   if (temp->numberOfAttributes(Dist_Stmt) > 0)
				dist_stmt->deleteStmt();
		   else
		   {
			dist_stmt->expr(0)->lhs()->setSymbol(temp);
if (!silentFlag)
printf("Change the distribute %s to template %s.\n",sym->identifier(),temp->identifier());

			  // modified Dist_Stmt attribute .
			sym->deleteAttribute(0,Dist_Stmt);
			temp->addAttribute(Dist_Stmt , dist_stmt ,sizeof(dist_stmt) );
		   }
		}
        }// end of for
}


/***************************************************************************
		is_align_root_distribute()
****************************************************************************/

void distribute_default(SgStatement *hedr_stmt , SgSymbol *symbol)
{
	int variant = hedr_stmt->variant();
	ASSERT(variant == PROG_HEDR || variant == PROC_HEDR || variant == FUNC_HEDR  || variant == MODULE_STMT);

	SgType *type=symbol->type();

	SgSymbol *proc;
        int dim=0;
        if (type->variant() == T_ARRAY)
                dim=((SgArrayType *)type)->dimension();

	int d ;
	if ( num_dims == -1)
		d = dim;
	else
		d = num_dims;		// if defined in argv
// check if already have PROCESSOR defined here .
	if ( hedr_stmt->numberOfAttributes(Proc_Def) > 0 ) {
		proc = (SgSymbol *)hedr_stmt->attributeValue(0,Proc_Def);
		SgArrayType *atype = isSgArrayType(proc->type());
		if (atype == NULL ) 
			d=0;
		else
		  	d = atype->dimension();
	}
	else
		while ((proc=get_dim_processor(hedr_stmt,d))==NULL && d>0 ) d--;
	if ( d==0 ) {
		printf("distribute_default: cannot distribute %s!\n",symbol->identifier());
		exit(1);
	}

	SgExpression *expr_list=NULL;
        for (int i=1; i<=dim ; i++)
        {
                SgExpression *ref ;
#ifdef DEFAULT_COLLPASE
			ref= new SgExpression(STAR_OP,NULL,NULL,NULL);
#else
		if (i>d) 
			ref= new SgExpression(STAR_OP,NULL,NULL,NULL);
		else
			ref= new SgExpression(BLOCK_OP,NULL,NULL,NULL);
#endif
                expr_list = new SgExpression(EXPR_LIST,ref,expr_list,NULL);
        }
	SgExpression *dist_format=new SgExpression(DIST_FORMAT_OP,NULL,expr_list,NULL);
	SgExpression *varref = new SgVarRefExp(proc);
	SgExpression *dist_target=new SgExpression(DIST_TARGET_OP,NULL,varref,NULL);
	SgExpression *dist_op=new SgExpression(DISTRIBUTE_OP,dist_format,dist_target,NULL);
	expr_list = new SgExpression(EXPR_LIST,dist_op,NULL,NULL);

	varref = new SgVarRefExp(symbol);
	SgExpression *var_list = new SgExpression(EXPR_LIST,varref,NULL,NULL);

	SgStatement *dist_decl=new SgStatement(DISTRIBUTE_DECL,NULL,NULL,var_list,NULL,expr_list);

	// insert statement
	SgStatement *decl = symbol->declareInStmt(hedr_stmt);
	if (decl==NULL) decl=hedr_stmt->firstStmtForDeclaration();
        decl->insertStmtAfter(*dist_decl,*hedr_stmt);
	// modify attributes
        symbol->addAttribute(Dist_Stmt , dist_decl ,sizeof(dist_decl) );
}

//  must be done last
// if has INHERIT attr , donot do
void is_align_root_distribute(SgStatement *hedr_stmt)
{
	for (int i=0 ; i< hedr_stmt->numberOfAttributes(Align_Root) ; i++)
        {
                SgSymbol *sym=(SgSymbol*)hedr_stmt->attributeValue(i,Align_Root);
		if (sym->numberOfAttributes(Dist_Stmt) > 0) continue;

		  // Dynamic attribute
		//if (sym->numberOfAttributes(Dynamic_Stmt) > 0) continue;
/*
		  // Allocatable
	        if (sym->attributes() &  ALLOCATABLE_BIT) continue;
		  // Pointer
	        if (sym->attributes() &  POINTER_BIT) continue;
*/

		SgType *atype =sym->type();
                if (atype->variant() == T_ARRAY ) atype=atype->baseType();
 	        if (atype->variant() == T_TEMPLATE)   // no align son 
		{
			if (sym->numberOfAttributes(Align_Son)==0 &&
			    sym->numberOfAttributes(Dynamic_Stmt)==0 ) 
			{
				  // then this template seems useless
if (!silentFlag)
printf("Delete an unused template : %s\n", sym->identifier());
				SgVarDeclStmt *stmt = (SgVarDeclStmt*)sym->attributeValue(0,Temp_Stmt);
				stmt->deleteTheSymbol(*sym);
				continue;
			}
		}
		  // if have INHERIT attribute
		if (sym->attributes() & INHERIT_BIT) continue;

		  // make a default distribute
if (!silentFlag)
printf("Align Root: %s is not distributed \n",sym->identifier());
		distribute_default(hedr_stmt,sym);
	}
}

